"""Script to run the gain selection over a list of dates."""
import logging
import re
import glob
import pandas as pd
import subprocess as sp
from pathlib import Path
from textwrap import dedent
import argparse
import sys

from astropy.table import Table
from lstchain.paths import parse_r0_filename
from datetime import datetime

from osa.scripts.reprocessing import get_list_of_dates, check_job_status_and_wait
from osa.utils.utils import wait_for_daytime, date_to_dir, date_to_iso
from osa.utils.logging import myLogger
from osa.utils.iofile import append_to_file
from osa.utils.cliopts import valid_date
from osa.job import get_sacct_output, run_sacct, job_finished_in_timeout
from osa.configs.config import cfg
from osa.paths import DEFAULT_CFG
from osa.nightsummary.nightsummary import run_summary_table


log = myLogger(logging.getLogger(__name__))

PATH = "PATH=/fefs/aswg/software/offline_dvr/bin:$PATH"

parser = argparse.ArgumentParser()
parser.add_argument(
        "--check",                                                                                       
        action="store_true",
        default=False,
        help="Check if any job failed",
)
parser.add_argument(
        "--no-queue-check",
        action="store_true",
        default=False,
        help="Do not wait until the number of jobs in the slurm queue is < 1500",
)
parser.add_argument(
        "-c",                                                                                            
        "--config",                                                                                      
        action="store",
        type=Path,
        default=DEFAULT_CFG,
        help="Configuration file",
)
parser.add_argument(                                                                                     
        "-d",                                                                                            
        "--date",                                                                                        
        default=None,
        type=valid_date,
        help="Night to apply the gain selection in YYYY-MM-DD format",
)                                                                                                        
parser.add_argument(                                                                                     
        "-l",                                                                                            
        "--dates-file",
        default=None,
        help="List of dates to apply the gain selection. The input file should list"
        "the dates in the format YYYY-MM-DD, one date per line.",
)                                                                                                       
parser.add_argument(                                                                                     
        "-s",                                                                                            
        "--start-time",
        type=int,
        default=10,
        help="Time to (re)start gain selection in HH format. Default is 10.",
)                                                                                                       
parser.add_argument(                                                                                     
        "-e",                                                                                            
        "--end-time",
        type=int,
        default=18,
        help="Time to stop gain selection in HH format. Default is 18.",
)
parser.add_argument(                                                                                     
        "-t",                                                                                            
        "--tool",
        type=str,
        default=None,
        help="Choose tool to apply the gain selection regardless the date. Possible options are: lst_dvr (by default used for dates "
        "previous to 2023-12-05) and lstchain_r0_to_r0g (by default used for dates later than 2023-12-05).",
)
parser.add_argument(                                                                                          
        "--simulate",
        action="store_true",
        default=False,
        help="Simulate launching of the gain selection script. Dry run.",
)
parser.add_argument(
        "-v",                                                                                      
        "--verbose",
        action="store_true",
        default=False,
        help="Activate debugging mode.",
)

def get_sbatch_script(
    run_id: str,
    subrun: str,
    input_file: Path,
    output_dir: Path,
    log_dir: Path,
    log_file: Path,
    ref_time: int,
    ref_counter: int,
    module: int,
    ref_source: str,
    tool: str
):
    """Build the sbatch job pilot script for running the gain selection."""
    mem_per_job = cfg.get("SLURM", "MEMSIZE_GAINSEL")
    slurm_account = cfg.get("SLURM", "ACCOUNT")
    sbatch_script = dedent(
            f"""\
        #!/bin/bash

        #SBATCH -D {log_dir}
        #SBATCH -o "gain_selection_{run_id:05d}_{subrun:04d}_%j.log"
        #SBATCH --job-name "gain_selection_{run_id:05d}"
        #SBATCH --partition=short,long
        #SBATCH --mem={mem_per_job}
        #SBATCH --account={slurm_account}
        """
        )
    
    if tool == "lst_dvr":
        sbatch_script += dedent(
            f"""   
        #SBATCH --export {PATH}

        lst_dvr {input_file} {output_dir} {ref_time} {ref_counter} {module} {ref_source}
        """
        )
        
    elif tool == "lstchain_r0_to_r0g":
        cmd = f"lstchain_r0_to_r0g --R0-file={input_file} --output-dir={output_dir} --log={log_file}"
        if not cfg.getboolean("lstchain", "use_ff_heuristic_gain_selection"): 
            cmd += " --no-flatfield-heuristic"
        sbatch_script += dedent(cmd)

    return sbatch_script


def launch_gainsel_for_data_run(
    date: datetime, run: Table, output_dir: Path, r0_dir: Path, log_dir: Path, tool: str, simulate: bool = False
    ):
    """
    Create the gain selection sbatch script and launch it for a given run.
    
    Runs from before 20231205 without UCTS or TIB info are directly copied to the final directory.
    Subruns that do not have four streams are also directly copied.
    """
    run_id = run["run_id"]
    ref_time = run["dragon_reference_time"]
    ref_counter = run["dragon_reference_counter"]
    module = run["dragon_reference_module_index"]
    ref_source = run["dragon_reference_source"].upper()

    files = glob.glob(f"{r0_dir}/LST-1.?.Run{run_id:05d}.????.fits.fz")
    subrun_numbers = [int(file[-12:-8]) for file in files]
    
    if tool == "lst_dvr" and ref_source not in ["UCTS", "TIB"]:
        input_files = r0_dir.glob(f"LST-1.?.Run{run_id:05d}.????.fits.fz")
        
        if is_run_already_copied(date, run_id):
            log.info(f"The R0 files corresponding to run {run_id} have already been copied to the R0G directory.")
        else:
            if not simulate:
                for file in input_files:
                    log.debug(
                        f"Run {run_id} does not have UCTS or TIB info, so gain selection cannot"
                        f"be applied. Copying directly the R0 files to {output_dir}."
                    )
                    sp.run(["cp", file, output_dir])

            else:
                log.info(
                    f"Run {run_id} does not have UCTS or TIB info, so gain selection cannot"
                    f"be applied. Simulate copy of the R0 files directly to {output_dir}."
                )

    else:
        n_subruns = max(subrun_numbers) 

        for subrun in range(n_subruns + 1):

            r0_files = glob.glob(f"{r0_dir}/LST-1.?.Run{run_id:05d}.{subrun:04d}.fits.fz")

            if len(r0_files) != 4:
                if not simulate and not is_run_already_copied(date, run_id):
                    log.debug(f"Run {run_id:05d}.{subrun:04d} does not have 4 streams of R0 files, so gain"
                        f"selection cannot be applied. Copying directly the R0 files to {output_dir}.")
                    for file in r0_files:
                        sp.run(["cp", file, output_dir])
                elif is_run_already_copied(date, run_id):
                    log.debug(f"Run {run_id:05d}.{subrun:04d} does not have 4 streams of R0 files. The R0 files"
                        f"have already been copied to {output_dir}.")
                elif simulate:
                    log.debug(f"Run {run_id:05d}.{subrun:04d} does not have 4 streams of R0 files, so gain"
                        f"selection cannot be applied. Simulate copy of the R0 files directly to {output_dir}.")

            else:
                history_file = log_dir / f"gain_selection_{run_id:05d}.{subrun:04d}.history"
                if history_file.exists():
                    if not simulate:
                        update_history_file(run_id, subrun, log_dir, history_file)

                    if history_file.read_text() == "":   # history_file is empty
                        log.debug(f"Gain selection is still running for run {run_id:05d}.{subrun:04d}")
                        continue
                    else:
                        gainsel_rc = history_file.read_text().splitlines()[-1][-1]
                        if gainsel_rc == "1":
                            job_id = get_last_job_id(run_id, subrun, log_dir)
                            if job_finished_in_timeout(job_id) and not simulate:
                                # Relaunch the job that finished in TIMEOUT
                                job_file = log_dir / f"gain_selection_{run_id:05d}.{subrun:04d}.sh"
                                sp.run(["sbatch", job_file], stdout=sp.PIPE, stderr=sp.STDOUT, check=True)
                            else:
                                log.warning(f"Gain selection failed for run {run_id:05d}.{subrun:04d}")
                        elif gainsel_rc == "0":
                            log.debug(f"Gain selection finished successfully for run {run_id:05d}.{subrun:04d}, "
                                        "no additional jobs will be submitted for this subrun.") 
                else:
                    log.debug(f"Creating and launching the gain selection sbatch script for subrun {run_id:05d}.{subrun:04d}")
                    if not simulate:
                        log_file = log_dir / f"r0_to_r0g_{run_id:05d}.{subrun:04d}.log"
                        job_file = log_dir / f"gain_selection_{run_id:05d}.{subrun:04d}.sh"
                        r0_files.sort()
                        with open(job_file, "w") as f:
                            f.write(
                                get_sbatch_script(
                                    run_id,
                                    subrun,
                                    r0_files[0],
                                    output_dir,
                                    log_dir,
                                    log_file,
                                    ref_time,
                                    ref_counter,
                                    module,
                                    ref_source,
                                    tool,
                                )
                            )

                        #submit job
                        history_file.touch()
                        sp.run(["sbatch", job_file], stdout=sp.PIPE, stderr=sp.STDOUT, check=True)


def apply_gain_selection(date: datetime, start: int, end: int, tool: str = None, no_queue_check: bool = False, simulate: bool = False):
    """
    Submit the jobs to apply the gain selection to the data for a given date
    on a subrun-by-subrun basis.
    """

    if not tool:
        if date_to_dir(date) < "20231205":
            tool = "lst_dvr"
        else:
            tool = "lstchain_r0_to_r0g"

    summary_table = run_summary_table(date)

    if len(summary_table) == 0:
        log.warning(f"No runs are found in the run summary of {date_to_iso(date)}. Nothing to do. Exiting.")
        sys.exit(0)

    # Apply gain selection only to DATA runs
    data_runs = summary_table[summary_table["run_type"] == "DATA"]
    log.info(f"Found {len(data_runs)} DATA runs to which apply the gain selection")

    date_str = date_to_dir(date)
    r0_dir = Path(cfg.get("LST1", "RAW_R0_DIR")) / date_str
    output_dir = Path(cfg.get("LST1", "R0_DIR")) / date_str
    log_dir = Path(cfg.get("LST1", "R0_DIR")) / f"log/{date_str}"
    if not simulate:
        output_dir.mkdir(parents=True, exist_ok=True)
        log_dir.mkdir(parents=True, exist_ok=True)

    for run in data_runs:
        if not no_queue_check:
            # Check slurm queue status and sleep for a while to avoid overwhelming the queue
            check_job_status_and_wait(max_jobs=1500)

        # Avoid running jobs while it is still night time
        wait_for_daytime(start, end)

        if not is_closed(date, run["run_id"]):
            launch_gainsel_for_data_run(date, run, output_dir, r0_dir, log_dir, tool, simulate)

    calib_runs = summary_table[summary_table["run_type"] != "DATA"]
    log.info(f"Found {len(calib_runs)} NO-DATA runs")

    for run in calib_runs:
        run_id = run["run_id"]
        
        if is_run_already_copied(date, run_id):
            log.info(f"The R0 files corresponding to run {run_id:05d} have already been copied, nothing to do.")
        else:
            log.info(f"Copying R0 files corresponding to run {run_id} directly to {output_dir}")
       	    if not simulate:
            	# Avoid copying files while it is still night time
            	wait_for_daytime(start, end)

            	r0_files = r0_dir.glob(f"LST-1.?.Run{run_id:05d}.????.fits.fz")

            	for file in r0_files:
                    sp.run(["cp", file, output_dir])


def get_last_job_id(run_id: str, subrun: str, log_dir: Path) -> int:
    """Get job id of the last gain selection job that was launched for a given subrun."""
    filenames = glob.glob(f"{log_dir}/gain_selection_{run_id:05d}_{subrun:04d}_*.log")
    if filenames:
        match = re.search(f'gain_selection_{run_id:05d}_{subrun:04d}_(\d+).log', sorted(filenames)[-1])
        job_id = match.group(1)
        return job_id


def update_history_file(run_id: str, subrun: str, log_dir: Path, history_file: Path):
    """
    Update the gain selection history file with the result 
    of the last job launched for a given subrun.
    """
    job_id = get_last_job_id(run_id, subrun, log_dir)
    if not job_id:
        log.debug(f"Cannot find a job_id for the run {run_id:05d}.{subrun:04d}")
    else:
        job_status = get_sacct_output(run_sacct(job_id=job_id))["State"]
        if job_status.item() in ["RUNNING", "PENDING"]:
            log.info(f"Job {job_id} is still running.")
            return
            
        elif job_status.item() == "COMPLETED":
            log.debug(f"Job {job_id} finished successfully, updating history file.")
            string_to_write = (
                f"{run_id:05d}.{subrun:04d} gain_selection 0\n"
            )
            append_to_file(history_file, string_to_write)
        
        else:
            log.info(f"Job {job_id} failed, updating history file.")
            string_to_write = (
                f"{run_id:05d}.{subrun:04d} gain_selection 1\n"
            )
            append_to_file(history_file, string_to_write)


def is_run_already_copied(date: datetime, run_id: int) -> bool:
    """Check if the R0 files of a given run have already been copied to the R0G directory."""
    r0_dir = Path(cfg.get("LST1", "RAW_R0_DIR"))
    r0g_dir = Path(cfg.get("LST1", "R0_DIR"))
    r0_files = glob.glob(f"{r0_dir}/{date_to_dir(date)}/LST-1.?.Run{run_id:05d}.????.fits.fz")
    r0g_files = glob.glob(f"{r0g_dir}/{date_to_dir(date)}/LST-1.?.Run{run_id:05d}.????.fits.fz")
    return len(r0_files)==len(r0g_files)


def is_closed(date: datetime, run_id: str) -> bool:
    """Check if run is already closed."""
    base_dir = Path(cfg.get("LST1", "BASE"))
    log_dir = base_dir / f"R0G/log/{date_to_dir(date)}"
    closed_run_file = log_dir / f"gain_selection_{run_id:05d}.closed"
    return closed_run_file.exists()


def GainSel_flag_file(date: datetime) -> Path:
    """Return the path to the file indicating the completion of the gain selection stage."""
    filename = cfg.get("LSTOSA", "gain_selection_check")
    GainSel_dir = Path(cfg.get("LST1", "GAIN_SELECTION_FLAG_DIR"))
    flagfile = GainSel_dir / date_to_dir(date) / filename
    return flagfile.resolve()


def GainSel_finished(date: datetime) -> bool:
    """Check if gain selection finished successfully."""
    flagfile = GainSel_flag_file(date)
    return flagfile.exists()
   

def check_gainsel_jobs_runwise(date: datetime, run_id: int) -> bool:
    """Search for failed jobs in the log directory."""
    base_dir = Path(cfg.get("LST1", "BASE"))
    log_dir = base_dir / f"R0G/log/{date_to_dir(date)}"
    history_files = list(log_dir.glob(f"gain_selection_{run_id:05d}.????.history"))
    summary_table = run_summary_table(date)
    n_subruns = summary_table[summary_table["run_id"] == run_id]["n_subruns"]
    
    if len(history_files) != n_subruns:
        log.debug(f"All history files of run {run_id} were not created yet")
        return False

    failed_subruns = []
    log.info(f"Checking all history files of run {run_id}")
    
    for file in history_files:
        match = re.search(f"gain_selection_{run_id:05d}.(\d+).history", str(file))
        subrun = match.group(1)
        if file.read_text() != "":
            gainsel_rc = file.read_text().splitlines()[-1][-1]

            if gainsel_rc == "1":
                log.warning(f"Gain selection failed for run {run_id}.{subrun}")
                failed_subruns.append(file)

            elif gainsel_rc == "0":
                log.debug(f"Gain selection finished successfully for run {run_id}.{subrun}")
        else:
            log.info(f"Gain selection is still running for run {run_id}.{subrun}")
            return False
            
    if failed_subruns:
        log.warning(f"{date_to_iso(date)}: Some gain selection jobs did not finish successfully for run {run_id}")
        return False
    else:
        log.info(f"{date_to_iso(date)}: All jobs finished successfully for run {run_id}, creating the corresponding .closed file")
        closed_run_file = log_dir / f"gain_selection_{run_id:05d}.closed"
        closed_run_file.touch()
        return True


def check_warnings_in_logs(date: datetime, run_id: int):
    """Look for warnings in the log files created by lstchain_r0_to_r0g."""
    base_dir = Path(cfg.get("LST1", "BASE"))
    log_dir = base_dir / f"R0G/log/{date_to_dir(date)}"
    log_files = log_dir.glob(f"r0_to_r0g_{run_id:05d}.*.log")
    for file in log_files:
        content = file.read_text().splitlines()
        for line in content:
            if "FlatField(FF)-like events are not tagged as FF" in line:
                log.warning(f"Warning for run {run_id}: {line}")


def check_failed_jobs(date: datetime):
    """Search for failed jobs in the log directory."""

    summary_table = run_summary_table(date)

    if len(summary_table) == 0:
        log.warning(f"No runs are found in the run summary of {date_to_iso(date)}. Nothing to do. Exiting.")
        sys.exit(0)
        
    data_runs = summary_table[summary_table["run_type"] == "DATA"]
    failed_runs = []

    for run in data_runs:
        run_id = run["run_id"]
        check_warnings_in_logs(date, run_id)
        if not is_closed(date, run_id):
            if not check_gainsel_jobs_runwise(date, run_id):
                log.warning(f"Gain selection did not finish successfully for run {run_id}.")
                failed_runs.append(run)

    if failed_runs:
        log.warning(f"Gain selection did not finish successfully for {date_to_iso(date)}, cannot create the flag file.")
        return

    runs = summary_table["run_id"]
    missing_runs = []

    date_str = date_to_dir(date)
    r0_dir = Path(cfg.get("LST1", "RAW_R0_DIR")) / date_str
    r0g_dir = Path(cfg.get("LST1", "R0_DIR")) / date_str
    r0_files = glob.glob(f"{r0_dir}/LST-1.?.Run?????.????.fits.fz")
    r0g_files = glob.glob(f"{r0g_dir}/LST-1.?.Run?????.????.fits.fz")
    all_r0_runs = [parse_r0_filename(i).run for i in r0_files]
    all_r0g_runs = [parse_r0_filename(i).run for i in r0g_files]

    for run in all_r0_runs:
        if run not in runs:
            if run not in all_r0g_runs:
                missing_runs.append(run)

    missing_runs.sort()
    if missing_runs:
        log.info(
            f"Some runs are missing. Copying R0 files of runs {pd.Series(missing_runs).unique()} "
            f"directly to {r0g_dir}"
        )

        for run in missing_runs:
            files = r0_dir.glob(f"LST-1.?.Run{run:05d}.????.fits.fz")
            for file in files:
                sp.run(["cp", file, r0g_dir])

    GainSel_dir = Path(cfg.get("LST1", "GAIN_SELECTION_FLAG_DIR"))
    flagfile_dir = GainSel_dir / date_str
    flagfile_dir.mkdir(parents=True, exist_ok=True)
    
    flagfile = GainSel_flag_file(date)
    log.info(f"Gain selection finished successfully, creating flag file for date {date_to_iso(date)} ({flagfile})")
    flagfile.touch()


def main():
    """
    Loop over the dates listed in the input file and launch the gain selection
    script for each of them. The input file should list the dates in the format
    YYYYMMDD one date per line.
    """
    args = parser.parse_args()
    
    if args.verbose:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)

    if args.date:
        if GainSel_finished(args.date):
            log.warning(f"Gain selection already done for date {date_to_iso(args.date)}. Exiting.")
            sys.exit(0)
        elif args.check:
            log.info(f"Checking gain selection status for date {date_to_iso(args.date)}")
            check_failed_jobs(args.date)
        else:
            log.info(f"\nApplying gain selection to date {date_to_iso(args.date)}")
            apply_gain_selection(
                args.date, 
                args.start_time, 
                args.end_time,
                args.tool,
                no_queue_check=args.no_queue_check, 
                simulate=args.simulate,
            )


    elif args.dates_file:
        list_of_dates = get_list_of_dates(args.dates_file)
        log.info(f"Found {len(list_of_dates)} dates to apply or check gain selection")

        if args.check:
            for date in list_of_dates:
                log.info(f"Checking gain selection status for date {date}")
                check_failed_jobs(date)
        else:
            for date in list_of_dates:
                log.info(f"Applying gain selection to date {date}")
                apply_gain_selection(
                    date, 
                    args.start_time, 
                    args.end_time,
                    args.tool,
                    no_queue_check=args.no_queue_check,
                    simulate=args.simulate,
                )
            log.info("Done! No more dates to process.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Script to copy analysis products to datacheck webserver creating new
directories whenever they are needed.
"""

import logging
from pathlib import Path

from osa.configs import options
from osa.configs.config import cfg
from osa.paths import (
    datacheck_directory,
    get_datacheck_files,
)
from osa.utils.cliopts import copy_datacheck_parsing
from osa.utils.logging import myLogger
from osa.utils.utils import DATACHECK_FILE_PATTERNS, date_to_dir
from osa.webserver.utils import copy_to_webserver

log = myLogger(logging.getLogger())


def are_files_copied(data_type: str, files: list) -> bool:
    """Check if all files of a given data type are copied."""
    n_files = len(files)

    if n_files == 0:
        log.warning(f"No {data_type} files found.")
        return False

    if data_type == "PEDESTAL" and n_files != 1:
        log.warning(f"Expected at least 1 PDF file, {n_files} found.")
        return False

    if data_type == "CALIB" and n_files != 1:
        log.warning(f"Expected at least 1 PDF file, {n_files} found.")
        return False

    if data_type == "DL1AB" and n_files != get_number_of_runs():
        log.warning(f"Expected {get_number_of_runs()} PDF files, {n_files} found.")
        return False

    if data_type == "LONGTERM" and n_files != 3:
        log.warning(f"Expected 3 DL1 check files (HTML, h5 and log), {n_files} found.")
        return False

    return True


def main():
    """Copy datacheck products to the webserver."""
    log.setLevel(logging.INFO)

    log.info(
        """
        ___________________________________________________________________________
        Expected PDF datacheck files: DRS4, ENF calibration & run-wise DL1.
        Additionally, daily DL1 log, HTML and h5 files should be copied.
        If any of these are missing it might happen that they are not produced yet.
        ___________________________________________________________________________
        """
    )

    copy_datacheck_parsing()
    nightdir = date_to_dir(options.date)

    all_files_are_copied = False

    for data_type, pattern in DATACHECK_FILE_PATTERNS.items():
        log.info(f"Looking for {pattern}")
        directory = datacheck_directory(data_type=data_type, date=nightdir)
        files = get_datacheck_files(pattern, directory)
        if len(files) != 0:
            copy_to_webserver(files, data_type, nightdir, options.prod_id)

        # Check if all files are copied
        all_files_are_copied = are_files_copied(data_type, files)

    if all_files_are_copied:
        log.info("All datacheck files copied. No more files are expected.")
    else:
        log.warning(
            "Not all datacheck files were copied. Check for problems or whether "
            "data were actually taken."
        )


def get_number_of_runs():
    """
    Get the run sequence processed list for the given date by globbing the
    run-wise DL1 files.
    """
    nightdir = date_to_dir(options.date)
    dl1_directory = Path(cfg.get("LST1", "DL1_DIR")) / nightdir / options.prod_id
    list_files = list(dl1_directory.glob("tailcut*/dl1_LST-1.Run?????.h5"))
    return len(list_files)


if __name__ == "__main__":
    main()



import glob
import re
import argparse
import logging
from pathlib import Path
from astropy.table import Table
import subprocess as sp

from osa.configs import options
from osa.configs.config import cfg
from osa.nightsummary.extract import get_last_pedcalib
from osa.utils.cliopts import valid_date, set_default_date_if_needed
from osa.utils.logging import myLogger
from osa.job import run_sacct, get_sacct_output
from osa.utils.utils import date_to_dir, get_calib_filters, get_lstchain_version
from osa.paths import (
    catB_closed_file_exists,
    catB_calibration_file_exists,
    analysis_path,
    get_major_version
)

log = myLogger(logging.getLogger())

parser = argparse.ArgumentParser()
parser.add_argument(
    "-c",                                                                                            
    "--config",                                                                                      
    action="store",
    type=Path,
    help="Configuration file",
)
parser.add_argument(                                                                                     
    "-d",                                                                                            
    "--date",
    type=valid_date,
    default=None,
)
parser.add_argument(
    "-v",                                                                                      
    "--verbose",
    action="store_true",
    default=False,
    help="Activate debugging mode.",
)
parser.add_argument(
    "-s",
    "--simulate",
    action="store_true",
    default=False,
    help="Simulate launching of the sequencer_catB_tailcuts script.",
)
parser.add_argument(
    "--overwrite-tailcuts",
    action="store_true",
    default=False,
    help="Overwrite the tailcuts config file if it already exists.",
)
parser.add_argument(
    "--overwrite-catB",
    action="store_true",
    default=False,
    help="Overwrite the Cat-B calibration files if they already exist.",
)
parser.add_argument(
    "tel_id",
    choices=["ST", "LST1", "LST2", "all"],
    help="telescope identifier LST1, LST2, ST or all.",
)

def are_all_history_files_created(run_id: int) -> bool:
    """Check if all the history files (one per subrun) were created for a given run."""
    run_summary_dir = Path(cfg.get(options.tel_id, "RUN_SUMMARY_DIR"))
    run_summary_file = run_summary_dir / f"RunSummary_{date_to_dir(options.date)}.ecsv"
    run_summary = Table.read(run_summary_file)
    n_subruns = run_summary[run_summary["run_id"] == run_id]["n_subruns"]
    analysis_dir = Path(options.directory)
    history_files = glob.glob(f"{str(analysis_dir)}/sequence_LST1_{run_id:05d}.????.history")
    if len(history_files) == n_subruns:
        return True
    else:
        return False


def r0_to_dl1_step_finished_for_run(run_id: int) -> bool:
    """
    Check if the step r0_to_dl1 finished successfully 
    for a given run by looking the history files.
    """
    if not are_all_history_files_created(run_id):
        log.debug(f"All history files for run {run_id:05d} were not created yet.")
        return False
    analysis_dir = Path(options.directory)
    history_files = glob.glob(f"{str(analysis_dir)}/sequence_LST1_{run_id:05d}.????.history")
    for file in history_files:
        rc = Path(file).read_text().splitlines()[-1][-1]
        if rc != "0":
            print(f"r0_to_dl1 step did not finish successfully (check file {file})")
            return False
    return True


def get_catB_last_job_id(run_id: int) -> int:
    """Get job id of the last Cat-B calibration job that was launched for a given run."""
    log_dir = Path(options.directory) / "log"
    filenames = glob.glob(f"{log_dir}/catB_calibration_{run_id:05d}_*.err")
    if filenames:
        match = re.search(f"catB_calibration_{run_id:05d}_(\d+).err", sorted(filenames)[-1])
        job_id = match.group(1)
        return job_id


def launch_catB_calibration(run_id: int):
    """
    Launch the Cat-B calibration script for a given run if the Cat-B calibration 
    file has not been created yet. If the Cat-B calibration script was launched
    before and it finished successfully, it creates a catB_{run}.closed file.
    """
    job_id = get_catB_last_job_id(run_id)
    if job_id and not options.overwrite_catB:
        job_status = get_sacct_output(run_sacct(job_id=job_id))["State"]
        if job_status.item() in ["RUNNING", "PENDING"]:
            log.debug(f"Job {job_id} (corresponding to run {run_id:05d}) is still running.")

        elif job_status.item() == "COMPLETED":
            catB_closed_file = Path(options.directory) / f"catB_{run_id:05d}.closed"
            catB_closed_file.touch()
            log.debug(
                f"Cat-B job {job_id} (corresponding to run {run_id:05d}) finished "
                f"successfully. Creating file {catB_closed_file}"
            )

        else: 
            log.warning(f"Cat-B job {job_id} (corresponding to run {run_id:05d}) failed.")

    else:
        if catB_calibration_file_exists(run_id):
            if not options.overwrite_catB:
                log.info(f"Cat-B calibration file already produced for run {run_id:05d}.")
                return 
            else:
                log.info(f"Cat-B calibration file already produced for run {run_id:05d}. Overwriting it.")

        command = cfg.get("lstchain", "catB_calibration")
        if cfg.getboolean("lstchain", "use_lstcam_env_for_CatB_calib"):
            env_command = f"conda run -n lstcam-env {command}"
        else:
            env_command = command
        options.filters = get_calib_filters(run_id) 
        base_dir = Path(cfg.get(options.tel_id, "BASE")).resolve()
        r0_dir = Path(cfg.get(options.tel_id, "R0_DIR")).resolve()
        log_dir = Path(options.directory) / "log"
        catA_calib_run = get_last_pedcalib(options.date)
        slurm_account = cfg.get("SLURM", "ACCOUNT")
        lstchain_version = get_major_version(get_lstchain_version())
        analysis_dir = cfg.get("LST1", "ANALYSIS_DIR")
        cmd = ["sbatch", f"--account={slurm_account}", "--parsable",
            "-o", f"{log_dir}/catB_calibration_{run_id:05d}_%j.out",
            "-e", f"{log_dir}/catB_calibration_{run_id:05d}_%j.err",
            env_command,
            f"-r {run_id:05d}",
            f"--catA_calibration_run={catA_calib_run}",
            "-b", base_dir,
            f"--r0-dir={r0_dir}",
            f"--filters={options.filters}",
        ]
        
        if command=="onsite_create_cat_B_calibration_file":
            cmd.append(f"--interleaved-dir={analysis_dir}")
        elif command=="lstcam_calib_onsite_create_cat_B_calibration_file":
            cmd.append(f"--dl1-dir={analysis_dir}")
            cmd.append(f"--lstchain-version={lstchain_version[1:]}")

        if options.overwrite_catB:
            cmd.append("--yes")

        if not options.simulate:
            job = sp.run(cmd, encoding="utf-8", capture_output=True, text=True, check=True)
            job_id = job.stdout.strip()
            log.debug(f"Launched Cat-B calibration job {job_id} for run {run_id}!")

        else: 
            log.info(f"Simulate launching of the {command} script.")
            

def launch_tailcuts_finder(run_id: int):
    """
    Launch the lstchain script to calculate the correct
    tailcuts to use for a given run. 
    """
    command = cfg.get("lstchain", "tailcuts_finder")
    slurm_account = cfg.get("SLURM", "ACCOUNT")
    input_dir = Path(options.directory)
    output_dir = Path(cfg.get(options.tel_id, "TAILCUTS_FINDER_DIR"))
    log_dir = Path(options.directory) / "log"
    log_file = log_dir / f"tailcuts_finder_{run_id:05d}_%j.log"
    cmd = [
        "sbatch", "--parsable",
        f"--account={slurm_account}",
        "-o", log_file,
        command,
        f"--input-dir={input_dir}",
        f"--run={run_id}",
        f"--output-dir={output_dir}",
    ]
    if not options.simulate:
        job = sp.run(cmd, encoding="utf-8", capture_output=True, text=True, check=True)
        job_id = job.stdout.strip()
        log.debug(f"Launched lstchain_find_tailcuts job {job_id} for run {run_id}!")

    else: 
        log.info(f"Simulate launching of the {command} script.")



def tailcuts_config_file_exists(run_id: int) -> bool:
    """Check if the config file created by the tailcuts finder script already exists."""
    tailcuts_config_file = Path(cfg.get(options.tel_id, "TAILCUTS_FINDER_DIR")) / f"dl1ab_Run{run_id:05d}.json"
    return tailcuts_config_file.exists()
    
        
def main():
    """
    Main script to be called as cron job. It launches the Cat-B calibration script 
    and the tailcuts finder script for each run of the corresponding date, and creates
    the catB_{run}.closed files if Cat-B calibration has finished successfully.
    """ 
    opts = parser.parse_args()
    options.tel_id = opts.tel_id
    options.simulate = opts.simulate
    options.overwrite_tailcuts = opts.overwrite_tailcuts
    options.overwrite_catB = opts.overwrite_catB
    options.date = opts.date
    options.date = set_default_date_if_needed()
    options.configfile = opts.config.resolve()
    options.directory = analysis_path(options.tel_id)

    if opts.verbose:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)

    run_summary_dir = Path(cfg.get(options.tel_id, "RUN_SUMMARY_DIR"))
    run_summary = Table.read(run_summary_dir / f"RunSummary_{date_to_dir(options.date)}.ecsv")
    data_runs = run_summary[run_summary["run_type"]=="DATA"]
    for run_id in data_runs["run_id"]:
        # first check if the dl1a files are produced
        if not r0_to_dl1_step_finished_for_run(run_id):
            log.info(f"The r0_to_dl1 step did not finish yet for run {run_id:05d}. Please try again later.")
        else:
            # launch catB calibration and tailcut finder in parallel
            if cfg.getboolean("lstchain", "apply_catB_calibration") and not catB_closed_file_exists(run_id):
                launch_catB_calibration(run_id)
            if not cfg.getboolean("lstchain", "apply_standard_dl1b_config"):
                if tailcuts_config_file_exists(run_id) and not options.overwrite_tailcuts:
                    log.debug(
                        f"Tailcuts config file already exists for run {run_id:05d}. Use --overwrite-tailcuts to overwrite it."
                    )
                else:
                    launch_tailcuts_finder(run_id)


if __name__ == "__main__":
    main()



"""Functions to deal with dates and prod IDs."""


import inspect
import logging
import os
import re
import json
import time
import numpy as np
from datetime import datetime, timedelta
from pathlib import Path
from socket import gethostname
import subprocess as sp
from gammapy.data import observatory_locations
from astropy import units as u
from astropy.table import Table
from lstchain.onsite import find_filter_wheels

import osa.paths
from osa.configs import options
from osa.configs.config import cfg
from osa.utils.iofile import write_to_file
from osa.utils.logging import myLogger


__all__ = [
    "get_lstchain_version",
    "date_to_dir",
    "date_to_iso",
    "is_day_closed",
    "get_prod_id",
    "night_finished_flag",
    "is_defined",
    "create_lock",
    "stringify",
    "gettag",
    "time_to_seconds",
    "DATACHECK_FILE_PATTERNS",
    "YESTERDAY",
    "is_night_time",
    "cron_lock",
    "example_seq",
    "wait_for_daytime",
]

log = myLogger(logging.getLogger(__name__))

DATACHECK_PRODUCTS = ["drs4", "enf_calibration", "dl1"]

DATACHECK_FILE_PATTERNS = {
    "PEDESTAL": "drs4*.pdf",
    "CALIB": "calibration*.pdf",
    "DL1AB": "datacheck_dl1*.pdf",
    "LONGTERM": "DL1_datacheck_*.*",
}

YESTERDAY = datetime.now() - timedelta(days=1)


def get_lstchain_version():
    """
    Get the lstchain version.

    Returns
    -------
    lstchain_version: string
    """
    from lstchain import __version__

    return f"v{__version__}"


def get_prod_id():
    """
    Get production ID from the configuration file if it is defined.
    Otherwise, it takes the lstchain version used.

    Returns
    -------
    prod_id: string
    """
    if not options.prod_id:
        if cfg.get("LST1", "PROD_ID") is not None:
            options.prod_id = cfg.get("LST1", "PROD_ID")
        else:
            options.prod_id = get_lstchain_version()

    log.debug(f"Getting prod ID for the running analysis directory: {options.prod_id}")

    return options.prod_id


def create_lock(lockfile) -> bool:
    """
    Create a lock file to prevent multiple instances of the same analysis.

    Parameters
    ----------
    lockfile: pathlib.Path

    Returns
    -------
    bool
    """
    directory_lock = lockfile.parent
    if options.simulate:
        log.debug(f"Simulate the creation of lock file {lockfile}")
    elif lockfile.exists() and lockfile.is_file():
        with open(lockfile, "r") as f:
            hostpid = f.readline()
        log.error(f"Lock by a previous process {hostpid}, exiting!\n")
        return True
    else:
        if not directory_lock.exists():
            directory_lock.mkdir(exist_ok=True, parents=True)
            log.debug(f"Creating parent directory {directory_lock} for lock file")
            pid = str(os.getpid())
            hostname = gethostname()
            content = f"{hostname}:{pid}"
            write_to_file(lockfile, content)
            log.debug(f"Lock file {lockfile} created")
            return True
    return False


def night_finished_flag() -> Path:
    """
    Create night-is-finished lock file.

    Returns
    -------
    lockfile: pathlib.Path
        Path of the lock file
    """
    basename = cfg.get("LSTOSA", "end_of_activity")
    date = date_to_dir(options.date)
    close_directory = Path(cfg.get(options.tel_id, "CLOSER_DIR"))
    lock_file = close_directory / date / options.prod_id / basename
    log.debug(f"Looking for lock file {lock_file}")
    return lock_file.resolve()


def date_to_iso(date: datetime) -> str:
    """Function to change from YYYY-MM-DD to YYYY-MM-DD."""
    return date.strftime("%Y-%m-%d")


def date_to_dir(date: datetime) -> str:
    """Function to change from YYYY-MM-DD to YYYYMMDD format (used for directories)."""
    return date.strftime("%Y%m%d")


def is_defined(variable):
    """Check if a variable is already defined."""
    try:
        variable
    except NameError:
        variable = None
    return variable is not None


def is_day_closed() -> bool:
    """Get the name and Check for the existence of the Closer flag file."""
    flag_file = night_finished_flag()
    return flag_file.exists()


def stringify(args):
    """Join a list of arguments in a string."""
    return " ".join(map(str, args))


def gettag():
    """Get the name of the script currently being used."""
    parent_file = os.path.basename(inspect.stack()[1][1])
    parent_module = inspect.stack()[1][3]
    return f"{parent_file}({parent_module})"


def time_to_seconds(timestring):
    """
    Transform (D-)HH:MM:SS time format to seconds.

    Parameters
    ----------
    timestring: str or None
        Time in format (D-)HH:MM:SS

    Returns
    -------
    Seconds that correspond to (D-)HH:MM:SS
    """
    if timestring is None:
        timestring = "00:00:00"
    if "-" in timestring:
        # Day is also specified (D-)HH:MM:SS
        days, hhmmss = timestring.split("-")
        hours, minutes, seconds = hhmmss.split(":")
        return int(days) * 24 * 3600 + int(hours) * 3600 + int(minutes) * 60 + int(seconds)

    split_time = timestring.split(":")
    if len(split_time) == 2:
        # MM:SS
        minutes, seconds = split_time
        hours = 0
    elif len(split_time) == 3:
        # HH:MM:SS
        hours, minutes, seconds = split_time
    else:
        raise ValueError("Time format not recognized.")
    return int(hours) * 3600 + int(minutes) * 60 + int(seconds)


def is_night_time(hour):
    """Check if it is nighttime."""
    if 8 <= hour <= 18:
        return False
    log.error("It is dark outside...")
    return True


def example_seq():
    """Example sequence table output for testing."""
    return "./extra/example_sequencer.txt"


def cron_lock(tel) -> Path:
    """Create a lock file for the cron jobs."""
    return osa.paths.analysis_path(tel) / "cron.lock"


def wait_for_daytime(start=8, end=18):
    """
    Check every hour if it is still nighttime
    to not running jobs while it is still night.
    """
    while time.localtime().tm_hour <= start or time.localtime().tm_hour >= end:
        log.info("Waiting for sunrise to not interfere with the data-taking. Sleeping.")
        time.sleep(3600)


def get_calib_filters(run_id):
    """Get the filters used for the calibration."""
    if options.test:  # Run tests avoiding the access to the database
        return 52

    else:
        mongodb = cfg.get("database", "caco_db")
        try:
            # Cast run_id to int to avoid problems with numpy int64 encoding in MongoDB
            return find_filter_wheels(int(run_id), mongodb)
        except IOError:
            log.warning("No filter information found in database. Assuming positions 52.")
            return 52


def culmination_angle(dec: u.Quantity) -> u.Quantity:  
    """
    Calculate culmination angle for a given declination.

    Parameters
    ----------
    dec: Quantity
        declination coordinate in degrees

    Returns
    -------
    Culmination angle in degrees
    """
    location = observatory_locations["cta_north"]
    Lat = location.lat  # latitude of the LST1 site    
    return abs(Lat - dec)


def convert_dec_string(dec_str: str) -> u.Quantity:
    """Return the declination angle in degrees corresponding to a 
    given string of the form "dec_XXXX" or "dec_min_XXXX"."""
    
    # Check if dec_str has a valid format
    pattern = r'^dec_(\d{3,4})$|^dec_min_(\d{3,4})$'
    if re.match(pattern, dec_str):
        
        # Split the string into parts
        parts = dec_str.split('_')

        # Extract the sign, degrees, and minutes
        sign = 1 if 'min' not in parts else -1
        degrees = int(parts[-1])

        # Calculate the numerical value
        dec_value = sign * (degrees / 100)

        return dec_value*u.deg


def get_declinations_dict(list1: list, list2: list) -> dict:
    """Return a dictionary created from two given lists."""
    corresponding_dict = {}
    for index, element in enumerate(list2):
        corresponding_dict[element] = list1[index]
    return corresponding_dict
        

def get_nsb_dict(rf_models_dir: Path, rf_models_prefix: str) -> dict:
    """Return a dictionary with the NSB level of the RF models and the path to each model."""
    rf_models = sorted(rf_models_dir.glob(f"{rf_models_prefix}*"))
    pattern = r"nsb_tuning_([\d.]+)"
    nsb_dict = {
        float(re.search(pattern, str(rf_model)).group(1)): rf_model
        for rf_model in rf_models if re.search(pattern, str(rf_model))
    }
    return nsb_dict


def get_mc_nsb_dir(run_id: int, rf_models_dir: Path) -> Path:
    """
    Return the path of the RF models directory with the NSB level 
    closest to that of the data for a given run.
    """
    additional_nsb = get_nsb_level(run_id)
    rf_models_prefix = cfg.get("lstchain", "mc_prod")
    nsb_dict = get_nsb_dict(rf_models_dir, rf_models_prefix)
    closest_nsb_value = min(nsb_dict.keys(), key=lambda x: abs(float(x) - additional_nsb))

    return nsb_dict[closest_nsb_value]


def get_nsb_level(run_id):
    """Choose the closest NSB among those that are processed with the same cleaning level."""
    tailcuts_finder_dir = Path(cfg.get(options.tel_id, "TAILCUTS_FINDER_DIR"))
    log_file = tailcuts_finder_dir / f"log_find_tailcuts_Run{run_id:05d}.log"
    with open(log_file, "r") as file:
        log_content = file.read()
    match = re.search(r"Additional NSB rate \(over dark MC\): ([\d.]+)", log_content)
    nsb = float(match.group(1))
    
    dl1b_config_filename = tailcuts_finder_dir / f"dl1ab_Run{run_id:05d}.json"
    with open(dl1b_config_filename) as json_file:
        dl1b_config = json.load(json_file)
    picture_th = dl1b_config["tailcuts_clean_with_pedestal_threshold"]["picture_thresh"]

    nsb_levels = np.array([0.00, 0.07, 0.14, 0.22, 0.38, 0.50, 0.81, 1.25, 1.76, 2.34])
    pth = np.array([8, 8, 8, 8, 10, 10, 12, 14, 16, 18])
    candidate_nsbs = nsb_levels[pth==picture_th]

    diff = abs(candidate_nsbs - nsb)
    return candidate_nsbs[np.argsort(diff)][0]


def get_RF_model(run_id: int) -> Path:
    """Get the path of the RF models to be used in the DL2 production for a given run.
    
    The choice of the models is based on the adequate additional NSB level
    and the proper declination line of the MC used for the training.
    """
    run_catalog_dir = Path(cfg.get(options.tel_id, "RUN_CATALOG"))
    run_catalog_file = run_catalog_dir / f"RunCatalog_{date_to_dir(options.date)}.ecsv"
    run_catalog = Table.read(run_catalog_file)
    pointing_dec = run_catalog[run_catalog["run_id"]==run_id]["source_dec"]*u.deg  
    # the "source_dec" given in the run catalogs is not actually the source declination, but the pointing declination
    pointing_culmination = culmination_angle(pointing_dec)

    rf_models_base_dir = Path(cfg.get(options.tel_id, "RF_MODELS"))
    rf_models_dir = get_mc_nsb_dir(run_id, rf_models_base_dir)
    dec_list = os.listdir(rf_models_dir)
    dec_list = [i for i in dec_list if i.startswith("dec")]

    # Convert each string in the list to numerical values
    dec_values = [convert_dec_string(dec) for dec in dec_list]
    dec_values = [dec for dec in dec_values if dec is not None]
    
    closest_declination = min(dec_values, key=lambda x: abs(x - pointing_dec))
    closest_dec_culmination = culmination_angle(closest_declination)
    
    lst_location = observatory_locations["cta_north"]
    lst_latitude = lst_location.lat  # latitude of the LST1 site    
    closest_lines = sorted(sorted(dec_values, key=lambda x: abs(x - lst_latitude))[:2])

    if pointing_dec < closest_lines[0] or pointing_dec > closest_lines[1]:
        # If the pointing declination is between the two MC lines closest to the latitude of 
        # the LST1 site, this check is not necessary.
        log.debug(
            f"The declination closest to {pointing_dec} is: {closest_declination}."
            "Checking if the culmination angle is larger than the one of the pointing."
        )
        while closest_dec_culmination > pointing_culmination:
            # If the culmination angle of the closest declination line is larger than for
            # the pointing declination, remove it from the declination lines list and
            # look for the second closest declination line.
            declinations_dict = get_declinations_dict(dec_list, dec_values)
            declination_str = declinations_dict[closest_declination]
            dec_values.remove(closest_declination)
            dec_list.remove(declination_str)
            closest_declination = min(dec_values, key=lambda x: abs(x - pointing_dec))
            closest_dec_culmination = culmination_angle(closest_declination)
    
    log.debug(f"The declination line to use for the DL2 production is: {closest_declination}")
    
    declinations_dict = get_declinations_dict(dec_list, dec_values)
    declination_str = declinations_dict[closest_declination]

    rf_model_path = rf_models_dir / declination_str

    return rf_model_path.resolve()


def get_lstcam_calib_version(env_path: Path) -> str:
    """Get the version of the lstcam_calib package installed in the given environment."""
    if options.test or options.simulate:
        return "0.1.1"
    python_exe = f"{str(env_path)}/bin/python"
    cmd = [python_exe, "-m", "pip", "show", "lstcam_calib"]
    result = sp.run(cmd, capture_output=True, text=True, check=True)
    for line in result.stdout.split('\n'):
        if line.startswith('Version:'):
            return line.split(':', 1)[1].strip()


import datetime
import os
import subprocess as sp
from pathlib import Path
from textwrap import dedent

import pytest
import yaml

from osa.configs import options
from osa.configs.config import cfg
from osa.scripts.closer import is_sequencer_successful, is_finished_check

ALL_SCRIPTS = [
    "sequencer",
    "closer",
    "copy_datacheck",
    "datasequence",
    "calibration_pipeline",
    "show_run_summary",
    "provprocess",
    "simulate_processing",
    "dl3_stage",
    "theta2_significance",
    "source_coordinates",
    "sequencer_webmaker",
    "gainsel_webmaker",
]

options.date = datetime.datetime.fromisoformat("2020-01-17")
options.tel_id = "LST1"
options.prod_id = "v0.1.0"
options.directory = "test_osa/test_files0/running_analysis/20200117/v0.1.0/"


def remove_provlog():
    log_file = Path("prov.log")
    if log_file.is_file():
        log_file.unlink()


def run_program(*args):
    result = sp.run(args, stdout=sp.PIPE, stderr=sp.STDOUT, encoding="utf-8", check=True)

    if result.returncode != 0:
        new_line = "\n"
        raise ValueError(
            f"Running {args[0]} failed with return code {result.returncode}, output: "
            f"{new_line.join(result.stdout)}"
        )

    return result


@pytest.mark.parametrize("script", ALL_SCRIPTS)
def test_all_help(script):
    """Test for all scripts if at least the help works."""
    run_program(script, "--help")


def test_simulate_processing(
    drs4_time_calibration_files,
    systematic_correction_files,
    run_summary_file,
    r0_data,
    merged_run_summary,
    drive_log,
    dl1b_config_files,
    tailcuts_log_files,
    rf_models,
    tailcuts_finder_dir,
):

    for file in drs4_time_calibration_files:
        assert file.exists()

    for file in systematic_correction_files:
        assert file.exists()

    for r0_file in r0_data:
        assert r0_file.exists()

    assert run_summary_file.exists()
    assert merged_run_summary.exists()
    assert drive_log.exists()
    assert rf_models[1].exists()
    assert dl1b_config_files[0].exists()
    assert tailcuts_log_files[0].exists()

    remove_provlog()
    rc = run_program("simulate_processing", "-p", "--force", "-d", "2020-01-17", "LST1")
    assert rc.returncode == 0

    prov_dl1_path = Path("./test_osa/test_files0/DL1/20200117/v0.1.0/tailcut84/log")
    prov_dl2_path = Path("./test_osa/test_files0/DL2/20200117/v0.1.0/tailcut84/nsb_tuning_0.14/log")
    prov_file_dl1 = prov_dl1_path / "calibration_to_dl1_01807_prov.log"
    prov_file_dl2 = prov_dl2_path / "calibration_to_dl2_01807_prov.log"
    json_file_dl1 = prov_dl1_path / "calibration_to_dl1_01807_prov.json"
    json_file_dl2 = prov_dl2_path / "calibration_to_dl2_01807_prov.json"
    pdf_file_dl1 = prov_dl1_path / "calibration_to_dl1_01807_prov.pdf"
    pdf_file_dl2 = prov_dl2_path / "calibration_to_dl2_01807_prov.pdf"

    assert prov_file_dl1.exists()
    assert prov_file_dl2.exists()
    assert pdf_file_dl1.exists()
    assert pdf_file_dl2.exists()

    with open(json_file_dl1) as file:
        dl1 = yaml.safe_load(file)
    assert len(dl1["entity"]) == 44
    assert len(dl1["activity"]) == 5
    assert len(dl1["used"]) == 15
    assert len(dl1["wasGeneratedBy"]) == 10

    with open(json_file_dl2) as file:
        dl2 = yaml.safe_load(file)
    assert len(dl2["entity"]) == 44
    assert len(dl2["activity"]) == 5
    assert len(dl2["used"]) == 15
    assert len(dl2["wasGeneratedBy"]) == 10

    rc = run_program("simulate_processing", "-p", "-d", "2020-01-17", "LST1")
    assert rc.returncode == 0

    remove_provlog()
    rc = run_program("simulate_processing", "-p", "-d", "2020-01-17", "LST1")
    assert rc.returncode == 0


def test_simulated_sequencer(
    drs4_time_calibration_files,
    systematic_correction_files,
    run_summary_file,
    run_catalog,
    r0_data,
    merged_run_summary,
    gain_selection_flag_file,
    dl1b_config_files,
    tailcuts_log_files,
    rf_models,
    dl2_merged,
):
    assert run_summary_file.exists()
    assert run_catalog.exists()
    assert gain_selection_flag_file.exists()

    for r0_file in r0_data:
        assert r0_file.exists()

    for file in drs4_time_calibration_files:
        assert file.exists()

    for file in systematic_correction_files:
        assert file.exists()

    for file in dl2_merged:
        assert file.exists()

    rc = run_program("sequencer", "-d", "2020-01-17", "--no-gainsel", "-s", "-t", "LST1")

    assert rc.returncode == 0
    now = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M")
    assert rc.stdout == dedent(
        f"""\
        =================================== Starting sequencer.py at {now} UTC for LST, Telescope: LST1, Date: 2020-01-17 ===================================
        Tel   Seq  Parent  Type      Run   Subruns  Source        Action  Tries  JobID  State  CPU_time  Exit  DL1%  MUONS%  DL1AB%  DATACHECK%  DL2%  
        LST1    1  None    PEDCALIB  1809  5        None          None    None   None   None   None      None  None  None    None    None        None  
        LST1    2       1  DATA      1807  11       Crab          None    None   None   None   None      None     0       0       0           0   100  
        LST1    3       1  DATA      1808  9        MadeUpSource  None    None   None   None   None      None     0       0       0           0   100  
        """  # noqa: E501
    )


def test_sequencer(sequence_file_list):
    for sequence_file in sequence_file_list:
        assert sequence_file.exists()


def test_autocloser(running_analysis_dir):
    result = run_program(
        "autocloser",
        "--date",
        "2020-01-17",
        "--test",
        "LST1",
    )
    assert os.path.exists(running_analysis_dir)
    assert result.stdout.split()[-1] == "Exit"


def test_closer(
    r0g_data,
    run_catalog,
    running_analysis_dir,
    test_observed_data,
    run_summary_file,
    drs4_time_calibration_files,
    systematic_correction_files,
    merged_run_summary,
    longterm_dir,
    longterm_link_latest_dir,
    daily_datacheck_dl1_files,
    dl1b_config_files,
    tailcuts_log_files,
    rf_models,
):
    # First assure that the end of night flag is not set and remove it otherwise
    night_finished_flag = Path(
        "./test_osa/test_files0/OSA/Closer/20200117/v0.1.0/NightFinished.txt"
    )
    if night_finished_flag.exists():
        night_finished_flag.unlink()

    for r0_file in r0g_data:
        assert r0_file.exists()
    for file in drs4_time_calibration_files:
        assert file.exists()
    for file in systematic_correction_files:
        assert file.exists()
    assert running_analysis_dir.exists()
    assert run_summary_file.exists()
    for obs_file in test_observed_data:
        assert obs_file.exists()
    assert merged_run_summary.exists()
    assert longterm_dir.exists()
    assert longterm_link_latest_dir.exists()
    for check_file in daily_datacheck_dl1_files:
        assert check_file.exists()
    assert rf_models[2].exists()

    run_program("closer", "-y", "-v", "-t", "-d", "2020-01-17", "LST1")
    closed_seq_file = running_analysis_dir / "sequence_LST1_01809.closed"

    # Check that files have been moved to their final destinations
    assert os.path.exists(
       "./test_osa/test_files0/DL1/20200117/v0.1.0/muons/muons_LST-1.Run01808.0011.fits"
    )
    assert os.path.exists(
       "./test_osa/test_files0/DL1/20200117/v0.1.0/interleaved/interleaved_LST-1.Run01808.0011.h5"
    )
    assert os.path.exists(
        "./test_osa/test_files0/DL1/20200117/v0.1.0/tailcut84/dl1_LST-1.Run01808.0011.h5"
    )
    assert os.path.exists(
        "./test_osa/test_files0/DL1/20200117/v0.1.0/tailcut84/datacheck/"
        "datacheck_dl1_LST-1.Run01808.0011.h5"
    )
    # Assert that the link to dl1 and muons files have been created
    assert os.path.islink(
        "./test_osa/test_files0/running_analysis/20200117/v0.1.0/muons_LST-1.Run01808.0011.fits"
    )
    assert os.path.islink(
        "./test_osa/test_files0/running_analysis/20200117/v0.1.0/dl1_LST-1.Run01808.0011.h5"
    )

    assert night_finished_flag.exists()
    assert closed_seq_file.exists()


def test_datasequence(
    running_analysis_dir,
    run_catalog,
    run_catalog_dir,
    rf_models_base_dir,
    rf_models,
    catB_closed_file,
    dl1b_config_files,
    tailcuts_log_files,
):
    drs4_file = "drs4_pedestal.Run00001.0000.fits"
    calib_file = "calibration.Run00002.0000.hdf5"
    timecalib_file = "time_calibration.Run00002.0000.hdf5"
    systematic_correction_file = "no_sys_corrected_calibration_scan_fit_20210514.0000.h5"
    drive_file = "DrivePosition_20200117.txt"
    runsummary_file = "RunSummary_20200117.ecsv"
    prod_id = "v0.1.0"
    run_number = "01807.0000"
    options.directory = running_analysis_dir

    assert run_catalog_dir.exists()
    assert run_catalog.exists()
    assert rf_models_base_dir.exists()
    assert rf_models[1].exists()
    assert catB_closed_file.exists()
    assert dl1b_config_files[0].exists()

    output = run_program(
        "datasequence",
        "--date=2020-01-17",
        "--simulate",
        f"--prod-id={prod_id}",
        f"--drs4-pedestal-file={drs4_file}",
        f"--pedcal-file={calib_file}",
        f"--time-calib-file={timecalib_file}",
        f"--systematic-correction-file={systematic_correction_file}",
        f"--drive-file={drive_file}",
        f"--run-summary={runsummary_file}",
        f"--dl1b-config={dl1b_config_files[0]}",
        "--dl1-prod-id=tailcut84",
        run_number,
        "LST1",
    )
    assert output.returncode == 0


def test_calibration_pipeline(running_analysis_dir):
    options.prod_id = "v0.1.0"
    drs4_run_number = "01804"
    pedcal_run_number = "01805"
    options.directory = running_analysis_dir

    output = run_program(
        "calibration_pipeline",
        "--date=2020-01-17",
        "--simulate",
        f"--prod-id={options.prod_id}",
        f"--drs4-pedestal-run={drs4_run_number}",
        f"--pedcal-run={pedcal_run_number}",
        "LST1",
    )
    assert output.returncode == 0


def test_is_sequencer_successful(
        run_summary,
        running_analysis_dir,
        dl1b_config_files,
        tailcuts_log_files,
        rf_models,
        merged_run_summary,
    ):
    assert merged_run_summary.exists()
    options.directory = running_analysis_dir
    options.test = True
    seq_tuple = is_finished_check(run_summary)
    options.test = False
    assert is_sequencer_successful(seq_tuple) is True


def test_drs4_pedestal_cmd(base_test_dir):
    from osa.scripts.calibration_pipeline import drs4_pedestal_command

    cmd = drs4_pedestal_command(drs4_pedestal_run_id="01804")
    r0_dir = base_test_dir / "R0G"
    expected_command = [
        cfg.get("lstchain", "drs4_baseline"),
        "-r",
        "01804",
        "-b",
        base_test_dir,
        f"--r0-dir={r0_dir}",
        "--no-progress",
    ]
    assert cmd == expected_command


def test_calibration_file_cmd(base_test_dir):
    from osa.scripts.calibration_pipeline import calibration_file_command

    cmd = calibration_file_command(drs4_pedestal_run_id="01804", pedcal_run_id="01809")
    r0_dir = base_test_dir / "R0G"
    expected_command = [
        cfg.get("lstchain", "charge_calibration"),
        "-p",
        "01804",
        "-r",
        "01809",
        "-b",
        base_test_dir,
        f"--r0-dir={r0_dir}",
    ]
    assert cmd == expected_command


def test_daily_longterm_cmd():
    from osa.scripts.closer import daily_longterm_cmd

    job_ids = ["12345", "54321"]
    cmd = daily_longterm_cmd(parent_job_ids=job_ids)
    slurm_account = cfg.get("SLURM", "ACCOUNT")

    expected_cmd = [
        "sbatch",
        "--parsable",
        f"--account={slurm_account}",
        "-D",
        options.directory,
        "-o",
        "log/longterm_daily_%j.log",
        "--dependency=afterok:12345,54321",
        "lstchain_longterm_dl1_check",
        "--input-dir=test_osa/test_files0/DL1/20200117/v0.1.0/tailcut84/datacheck",
        "--output-file=test_osa/test_files0/OSA/DL1DataCheck_LongTerm/v0.1.0/20200117/DL1_datacheck_20200117.h5",
        "--muons-dir=test_osa/test_files0/DL1/20200117/v0.1.0/muons",
        "--batch",
    ]

    assert cmd == expected_cmd


def test_observation_finished():
    """Check if observation is finished for `options.date=2020-01-17`."""
    from osa.scripts.closer import observation_finished

    date1 = datetime.datetime(2020, 1, 21, 12, 0, 0)
    assert observation_finished(date=date1) is True
    date2 = datetime.datetime(2020, 1, 17, 5, 0, 0)
    assert observation_finished(date=date2) is False


def test_no_runs_found():
    output = sp.run(
        ["sequencer", "-s", "-d", "2015-01-01", "LST1"], text=True, stdout=sp.PIPE, stderr=sp.PIPE
    )
    assert output.returncode == 0
    assert "No runs found for this date. Nothing to do. Exiting." in output.stderr.splitlines()[-1]


@pytest.mark.skip(reason="Currently not working with all combinations")
def test_sequencer_webmaker(
    run_summary,
    merged_run_summary,
    drs4_time_calibration_files,
    systematic_correction_files,
    base_test_dir,
):
    # Check if night finished flag is set
    night_finished = base_test_dir / "OSA/Closer/20200117/v0.1.0/NightFinished.txt"

    if night_finished.exists():
        output = sp.run(
            ["sequencer_webmaker", "--test", "-d", "2020-01-17"],
            text=True,
            stdout=sp.PIPE,
            stderr=sp.PIPE,
        )
        assert output.returncode != 0
        assert output.stderr.splitlines()[-1] == "Date 2020-01-17 is already closed for LST1"
        night_finished.unlink()

    output = sp.run(["sequencer_webmaker", "--test", "-d", "2020-01-17"])
    assert output.returncode == 0
    directory = base_test_dir / "OSA" / "SequencerWeb"
    directory.mkdir(parents=True, exist_ok=True)
    expected_file = directory / "osa_status_20200117.html"
    assert expected_file.exists()

    output = sp.run(["sequencer_webmaker", "--test"])
    assert output.returncode != 0

    # Running without test option will make the script fail
    output = sp.run(["sequencer_webmaker", "-d", "2020-01-17"])
    assert output.returncode != 0


def test_gainsel_webmaker(
    base_test_dir,
):

    output = sp.run(["gainsel_webmaker", "-d", "2020-01-17"])
    assert output.returncode == 0
    directory = base_test_dir / "OSA" / "GainSelWeb"
    expected_file = directory / "osa_gainsel_status_2020-01-17.html"
    assert expected_file.exists()

    # Test a date with non-existing run summary
    output = sp.run(["gainsel_webmaker", "-d", "2024-01-12"])
    assert output.returncode == 0
    directory = base_test_dir / "OSA" / "GainSelWeb"
    expected_file = directory / "osa_gainsel_status_2024-01-12.html"
    assert expected_file.exists()


def test_gainsel_web_content():
    from osa.scripts.gainsel_webmaker import check_failed_jobs

    table = check_failed_jobs(options.date)
    assert table["GainSelStatus"][0] == "NOT STARTED"
    assert table["GainSel%"][0] == 0.0



import logging
from argparse import ArgumentParser
from datetime import datetime, timedelta
from pathlib import Path

import pandas as pd
from astropy.table import Table

from osa.configs import options
from osa.configs.config import cfg
from osa.nightsummary.nightsummary import run_summary_table
from osa.paths import DEFAULT_CFG
from osa.scripts.sequencer_webmaker import html_content
from osa.utils.utils import date_to_dir, date_to_iso

log = logging.getLogger(__name__)


def valid_date(string):
    """Check if the string is a valid date and return a datetime object."""
    return datetime.strptime(string, "%Y-%m-%d")


common_parser = ArgumentParser(add_help=False)
common_parser.add_argument(
    "-c",
    "--config",
    type=Path,
    default=DEFAULT_CFG,
    help="Use specific config file [default configs/sequencer.cfg]",
)
common_parser.add_argument(
    "-d",
    "--date",
    help="Date of the start of the night in ISO format (YYYY-MM-DD). Defaults to yesterday",
    type=valid_date,
)


def check_gainsel_jobs_runwise(date: datetime, run_id: int) -> bool:
    """Search for failed jobs in the log directory."""
    base_dir = Path(cfg.get("LST1", "BASE"))
    flat_date = date_to_dir(date)
    log_dir = base_dir / f"R0G/log/{flat_date}"
    history_files = log_dir.glob(f"gain_selection_{run_id:05d}.????.history")

    success_subruns = 0
    failed_subruns = 0
    pending_subruns = 0

    for file in history_files:
        if file.read_text() != "":
            gainsel_rc = file.read_text().splitlines()[-1][-1]

            if gainsel_rc == "1":
                failed_subruns += 1

            elif gainsel_rc == "0":
                success_subruns += 1

        else:
            pending_subruns += 1

    return {"pending": pending_subruns, "success": success_subruns, "failed": failed_subruns}


def check_failed_jobs(date: datetime) -> pd.DataFrame:
    """Search for failed jobs in the log directory."""
    summary_table = run_summary_table(date)
    data_runs = summary_table[summary_table["run_type"] == "DATA"]

    gainsel_status_dict = {}
    for run in data_runs:
        run_id = run["run_id"]
        gainsel_job_status = check_gainsel_jobs_runwise(date, run_id)
        gainsel_status_dict[run_id] = gainsel_job_status

    gainsel_df = pd.DataFrame(gainsel_status_dict.values(), index=gainsel_status_dict.keys())
    gainsel_df.reset_index(inplace=True)
    gainsel_df.rename(columns={"index": "run_id"}, inplace=True)
    summary_table = summary_table.to_pandas()

    final_table = pd.merge(summary_table, gainsel_df, on="run_id")[
        [
            "run_id",
            "n_subruns",
            "pending",
            "success",
            "failed",
        ]
    ]

    def determine_status(row):
        if row["failed"] > 0:
            return "FAILED"
        elif row["pending"] == row["n_subruns"]:
            return "PENDING"
        elif row["success"] == row["n_subruns"]:
            return "COMPLETED"
        elif row["pending"] > 0:
            return "RUNNING"
        else:
            return "NOT STARTED"

    final_table["GainSel%"] = round(final_table["success"] * 100 / final_table["n_subruns"])
    final_table["GainSelStatus"] = final_table.apply(determine_status, axis=1)

    return final_table


def main():
    """Produce the html file with the processing OSA Gain Selection status.

    It creates an HTML file osa_gainsel_status_YYYY-MM-DD.html
    """
    args = ArgumentParser(
        description=(
            "Script to create an HTML file with the gain selection status "
            "(osa_gainsel_status_YYYY-MM-DD.html)"
        ),
        parents=[common_parser],
    ).parse_args()

    if args.date:
        flat_date = date_to_dir(args.date)
        options.date = args.date

    else:
        # yesterday by default
        yesterday = datetime.now() - timedelta(days=1)
        options.date = yesterday
        flat_date = date_to_dir(yesterday)

    date = date_to_iso(options.date)
    run_summary_directory = Path(cfg.get("LST1", "RUN_SUMMARY_DIR"))
    run_summary_file = run_summary_directory / f"RunSummary_{flat_date}.ecsv"

    gain_selection_web_directory = Path(cfg.get("LST1", "GAIN_SELECTION_WEB_DIR"))
    gain_selection_web_directory.mkdir(parents=True, exist_ok=True)
    html_file = gain_selection_web_directory / f"osa_gainsel_status_{date}.html"

    # Create and save the HTML file
    if not run_summary_file.is_file() or len(Table.read(run_summary_file)["run_id"]) == 0:
        content = "<p>No data found</p>"
        log.warning(f"No data found for date {date}, creating an empty HTML file.")

    else:
        # Get the table with the gain selection check report in HTML format:
        table_gain_selection_jobs = check_failed_jobs(options.date)
        content = table_gain_selection_jobs.to_html(justify="left")

    html_file.write_text(html_content(content, date, "OSA Gain Selection"))
    log.info(f"Created HTML file {html_file}")


if __name__ == "__main__":
    main()
